/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <math.h>

#include "UsbPVCamDriver.h"
#include "drivers/legacy/status/GateStatus.h"
#include "drivers/legacy/status/DetectorStatus.h"

#include "UsbPVCamState.h"
#include "RealUsbPVCamDriver.h"
#include "PerfectUsbPVCamDriver.h"
#include "UsbPVCamDef.h"
#include "drivers/legacy/def/CommandDef.h"
#include "drivers/legacy/def/GateDef.h"
#include "drivers/legacy/def/DetectorDef.h"

// Trace include



using namespace boost;

/*
 * Constructor
 */
UsbPVCamDriver::UsbPVCamDriver(const std::string& name) :
	driver::Gate(name), driver::DetectorController(name), m_StartActivated(false),
			m_DataSize(0), m_MaskSize(0), m_Data(NULL), m_DataShort(NULL), m_Mask(NULL),
			m_SynchroniseReadActivated(false), m_SynchroniseReadTerminated(
					false) {
	registerParentFunction(NONE_FUNCTION);
	registerStates(new RealUsbPVCamDriver(this), new PerfectUsbPVCamDriver(this), new PerfectUsbPVCamDriver(this));

	registerCategory(driver::Gate::TYPE);
	registerCategory(driver::DetectorController::TYPE);

	registerStatus(new GateStatus());
	registerStatus(new DetectorStatus());

	/*
	 * Init the device command list
	 */
	initCommand(COMMAND_INIT_DEVICE_CONTAINER);
	initCommand(COMMAND_CLEAR_DEVICE_CONTAINER);
	initCommand(COMMAND_WRITE_DEVICE_CONTAINER);
	initCommand(COMMAND_WRITE_PARAM_DEVICE_CONTAINER);
	initCommand(COMMAND_READ_DEVICE_CONTAINER);
	initCommand(COMMAND_SYNCHRONISE_READ_DEVICE_CONTAINER);
	initCommand(COMMAND_START_DEVICE_CONTAINER);
	initCommand(COMMAND_CONTINUE_DEVICE_CONTAINER);
	initCommand(COMMAND_PAUSE_DEVICE_CONTAINER);
	initCommand(COMMAND_STOP_DEVICE_CONTAINER);
	initCommand(COMMAND_STATUS_DEVICE_CONTAINER);
	initCommand(COMMAND_READ_INFOS_DEVICE_CONTAINER);

	// Init functions
	registerFunction(NONE_FUNCTION);

	// Set the type of device
	updateValue(DEVICE_TYPE_DEVICE_CONTAINER, std::string(
			ROOT_DEVICE_TYPE_DEVICE_CONTAINER));

	/*
	 * Register the Spy and Observer commands necessary to do the updates.
	 */
	//registerObserverCommand(COMMAND_READ_DEVICE_CONTAINER, 100);
	registerObserverCommand(COMMAND_STATUS_DEVICE_CONTAINER, 50);

}

/*
 * Destructor
 */
UsbPVCamDriver::~UsbPVCamDriver() {
	if (m_Data != NULL) {
		delete[] m_Data;
	}
	if (m_Mask != NULL) {
		delete[] m_Mask;
	}
}

/*
 * execute
 */
void UsbPVCamDriver::execute(const std::string& aCommand) {
	Trace::getInstance()->trace("DeviceContainer", "UsbPVCamDriver", "execute",
			Level::s_Info, "Command = %s,", aCommand.c_str());

	if ((aCommand != COMMAND_PAUSE_DEVICE_CONTAINER) && (aCommand
			!= COMMAND_STOP_DEVICE_CONTAINER) && (aCommand
			!= COMMAND_STATUS_DEVICE_CONTAINER) && (aCommand
			!= COMMAND_READ_DEVICE_CONTAINER)) {
		sendProgressEvent(PROGRESSION_UNKNOWNSTATE_DEVICE_CONTAINER);
	}

	UsbPVCamState* currentState =
			static_cast<UsbPVCamState*> (getCurrentState());
	// Check command
	if (aCommand == COMMAND_INIT_DEVICE_CONTAINER) {
		// Init command
		currentState->init();
	} else if (aCommand == COMMAND_CLEAR_DEVICE_CONTAINER) {
		// Move command
		updateValue(DETECTOR_SUM_DEVICE_CONTAINER, (float64) 0.);
		currentState->clear();
	} else if (aCommand == COMMAND_WRITE_DEVICE_CONTAINER) {
		// Write command
		currentState->write();
	} else if (aCommand == COMMAND_WRITE_PARAM_DEVICE_CONTAINER) {
		// Write command
		currentState->writeParam();
	} else if (aCommand == COMMAND_READ_DEVICE_CONTAINER) {
		// Read command
		currentState->read();
	} else if (aCommand == COMMAND_SYNCHRONISE_READ_DEVICE_CONTAINER) {
		// Read command
		m_SynchroniseReadActivated = true;
		m_SynchroniseReadTerminated = false;
		currentState->synchroniseRead();
	} else if (aCommand == COMMAND_START_DEVICE_CONTAINER) {
		// Start command
		m_StartActivated = true;
		currentState->start();
	} else if (aCommand == COMMAND_CONTINUE_DEVICE_CONTAINER) {
		// Continue command
		currentState->continu();
	} else if (aCommand == COMMAND_PAUSE_DEVICE_CONTAINER) {
		// Stop command
		currentState->pause();
	} else if (aCommand == COMMAND_STOP_DEVICE_CONTAINER) {
		// Stop command
		currentState->stop();
		m_StartActivated = false;
	} else if (aCommand == COMMAND_STATUS_DEVICE_CONTAINER) {
		// Status command
		currentState->readStatus();
	} else if (aCommand == COMMAND_READ_INFOS_DEVICE_CONTAINER) {
		// Info command
		currentState->readInfos();
	} else {
		//Error bad command
		//Todo , normaly it will be detect it device action driver
	}
}

/*
 * refreshSetValue
 */
void UsbPVCamDriver::refreshSetValue(std::string propertyName,
		const boost::any aValue) throw (CannotSetValue) {
	/*
	 * Gate mode
	 */
	if (m_StartActivated == true) {
		throw CannotSetValue();
		// Test the division must be 10 multiple
	} else if (propertyName == GATE_MODE_DEVICE_CONTAINER) {
		if (any_cast<int32> (aValue) != (int32 ) GATE_TIME_MODE_DEVICE_CONTAINER) {
			throw CannotSetValue();
		}
	} else if (propertyName == GATE_ACTUAL_TIME_DEVICE_CONTAINER) {
		throw CannotSetValue();
	} else if (propertyName == GATE_ACTUAL_COUNTER1_DEVICE_CONTAINER) {
		throw CannotSetValue();
	} else if (propertyName == GATE_ACTUAL_COUNTER2_DEVICE_CONTAINER) {
		throw CannotSetValue();
	} else if (propertyName == GATE_ACTUAL_COUNTER3_DEVICE_CONTAINER) {
		throw CannotSetValue();
	} else if (propertyName == GATE_STATUS_DEVICE_CONTAINER) {
		throw CannotSetValue();
	}
	/*
	 * Detector mode
	 */
	else if (propertyName == DETECTOR_WANTED_MODE_DEVICE_CONTAINER) {
		if (any_cast<int32> (aValue)
				!= static_cast<int32> (DETECTOR_NO_TOF_DEVICE_CONTAINER)) {
			throw CannotSetValue();
		}
	} else if (propertyName == DETECTOR_DATA_DEVICE_CONTAINER) {
		throw CannotSetValue();
	} else if (propertyName == DETECTOR_SUM_DEVICE_CONTAINER) {
		throw CannotSetValue();
	} else if (propertyName == DETECTOR_MASK_DEVICE_CONTAINER) {
		throw CannotSetValue();
	} else if (propertyName == DETECTOR_KINE_TABLE_DEVICE_CONTAINER) {
		throw CannotSetValue();
	} else if (propertyName == DETECTOR_STATUS_DEVICE_CONTAINER) {
		throw CannotSetValue();
	} else if (propertyName == DETECTOR_ACTUAL_MODE_DEVICE_CONTAINER) {
		throw CannotSetValue();
	} else if (propertyName == DETECTOR_ACTUAL_CHANNEL_DEVICE_CONTAINER) {
		throw CannotSetValue();
	} else if (propertyName == DETECTOR_ACTUAL_DELAY_DEVICE_CONTAINER) {
		throw CannotSetValue();
	} else if (propertyName == DETECTOR_ACTUAL_DURATION_DEVICE_CONTAINER) {
		throw CannotSetValue();
	} else if (propertyName == DETECTOR_ACTUAL_NCHANNEL_DEVICE_CONTAINER) {
		throw CannotSetValue();
	} else if (propertyName == DETECTOR_DATA_SIZE_DEVICE_CONTAINER) {
		int32 newSize = any_cast<int32> (aValue);
		if (m_DataSize != newSize) {
			/*
			 * delete the current data array and create a new one
			 */

			if (m_Data != NULL)
				delete[] m_Data;
			m_Data = new int32[newSize];
			if (m_Data == NULL) {
				sendErrorEvent(NOTENOUGH_MEMORY_ERROR,
						"Can't create data table");
				throw CannotSetValue();
			}

			if (m_DataShort != NULL)
				delete[] m_DataShort;
			m_DataShort = new uint16[newSize];
			if (m_DataShort == NULL) {
				sendErrorEvent(NOTENOUGH_MEMORY_ERROR,
						"Can't create data table");
				throw CannotSetValue();
			}

			/*
			 * zero all the data
			 */
			memset(m_Data, 0, newSize * sizeof(int32 ));
			memset(m_DataShort, 0, newSize * sizeof(uint16));

			/*
			 * put the new array into the Property and set the size correctly
			 * May sure that the SetPoint and ActualValues point to the same array
			 */
			updateValue(DETECTOR_DATA_DEVICE_CONTAINER, m_Data);
			setSize(DETECTOR_DATA_DEVICE_CONTAINER, newSize);
			m_DataSize = newSize;
		}
	} else if (propertyName == DETECTOR_MASK_SIZE_DEVICE_CONTAINER) {
		int32 newSize = any_cast<int32> (aValue);
		if (m_MaskSize != newSize) {
			/*
			 * delete the current data array and create a new one
			 */
			if (m_Mask != NULL)
				delete[] m_Mask;
			m_Mask = new int32[newSize];
			if (m_Mask == NULL) {
				sendErrorEvent(NOTENOUGH_MEMORY_ERROR,
						"Can't create mask table");
				throw CannotSetValue();
			}

			/*
			 * zero all the data
			 */
			memset(m_Mask, 0, newSize * sizeof(int32 ));

			/*
			 * put the new array into the Property and set the size correctly
			 * May sure that the SetPoint and ActualValues point to the same array
			 */
			updateValue(DETECTOR_MASK_DEVICE_CONTAINER, m_Mask);
			setSize(DETECTOR_MASK_DEVICE_CONTAINER, newSize);
			m_MaskSize = newSize;
		}
	} else {
		DeviceDriver::refreshSetValue(propertyName, aValue);
	}
}

/*
 * calculateProgression
 */
void UsbPVCamDriver::calculateProgression() {
	int32 progression = 0;
	float64 setpoint = 0.0;
	float64 actual = 0.0;

	setpoint = any_cast<float64> (getValue(GATE_WANTED_TIME_DEVICE_CONTAINER));
	actual = any_cast<float64> (getValue(GATE_ACTUAL_TIME_DEVICE_CONTAINER));

	if (setpoint == 0.0) {
		sendProgressEvent(PROGRESSION_END_DEVICE_CONTAINER);
	} else {
		progression = (int32 ) (actual / setpoint
				* PROGRESSION_RATIO_DEVICE_CONTAINER);
		// Be sure that progression > 0%
		if (progression < PROGRESSION_UNKNOWNSTATE_DEVICE_CONTAINER)
			progression = 0;
		// Be sure that only Phased status put 100%
		if (progression >= PROGRESSION_END_DEVICE_CONTAINER)
			progression = PROGRESSION_END_DEVICE_CONTAINER - 1;
		sendProgressEvent(progression);
	}
}


const std::string UsbPVCamDriver::TYPE = "UsbPVCam";
